import pandas as pd
from zigzag import peak_valley_pivots
import numpy as np
import volprofile as vp
from gstargets.config import DIRECTION, Develop


def _fractional_candle(
    fraction=0.5, _dict={"open": None, "high": None, "low": None, "close": None}
):
    shadow = _dict["high"] - _dict["low"]
    body = abs(_dict["open"] - _dict["close"])
    if shadow == 0:
        return False
    return body / shadow < fraction


def _findWave(pivots, waveNum, direction):
    """
    get the wave relative to the end of chart for example if waveNum = 1 direction='up'
    get the first wave which the direction was upward
    it follows the notation of zigzag indicator which always ends with a dummy pivot
    """
    # validation check
    if waveNum < 1 or type(waveNum) != int:
        raise ValueError("not a valid waive number")
    if direction not in (DIRECTION.UP, DIRECTION.DOWN):
        raise ValueError("not a valid direction")

    # all trend changing indices
    indices = np.where(pivots != 0)[0]
    # iterate over the first to just one before the last one
    count, idx = 0, len(indices) - 2
    while count < waveNum and idx != 0:
        _index = indices[idx]
        _currentDir = DIRECTION.UP if pivots[_index] == 1 else DIRECTION.DOWN
        if direction == _currentDir:
            count += 1
        idx -= 1
    return indices[idx], indices[idx + 1]


def _get_min_idxs(volprofile_result):
    idxs = np.where(
        volprofile_result.aggregateVolume == np.min(volprofile_result.aggregateVolume)
    )[0]

    answers = []
    for idx in idxs:
        answers.append(int(volprofile_result.iloc[idx, :]["index"]))
    return answers


def _get_max_idxs(volprofile_result):
    idxs = list(np.where(
        volprofile_result.aggregateVolume == np.max(volprofile_result.aggregateVolume)
    )[0])

    print(idxs)

    volprofile_result['average'] = volprofile_result["aggregateVolume"].mean()
    volprofile_result["high"] = (volprofile_result.aggregateVolume - volprofile_result.average) / volprofile_result.average > 0.2 
    toadd = list(volprofile_result.loc[volprofile_result["high"]].index)
    idxs.extend(toadd)
    return idxs 


def _get_jumped_idxs(volprofile_result: pd.DataFrame, thresholdType2, tradeSide):
    step = 1 if tradeSide == DIRECTION.UP else -1
    answers = []
    Half = False
    prevBin = {"aggregateVolume": 0}
    for _, row in volprofile_result[::step].iterrows():
        if not row["valid"]:
            continue
        if row["aggregateVolume"] < prevBin["aggregateVolume"] * thresholdType2:
            Half = True
        elif Half:
            answers.append(prevBin["index"])
            Half = False
        prevBin = row

    return answers


def _get_high_volume_levels(
    volprofile_result: pd.DataFrame,
    current_price,
    trade_side,
):
    if trade_side == DIRECTION.DOWN:
        volprofile_result[volprofile_result.minPrice > current_price]["valid"] &= True
    else:
        volprofile_result[volprofile_result.maxPrice < current_price]["valid"] &= True

    _maxes = _get_max_idxs(volprofile_result[volprofile_result["valid"]])
    areas = [volprofile_result.iloc[_max, :].to_dict() for _max in _maxes]

    sorted_areas = sorted(areas, key=lambda area: area["minPrice"]) 
    to_combine = sorted_areas[0]
    finals = []
    for i in range(1, len(sorted_areas)):
        new = sorted_areas[i]
        if to_combine is None:
            to_combine = new
            continue
        if _can_combine(to_combine, new):
            to_combine = _get_combined(to_combine, new)
            continue
        else:
            finals.append(to_combine)
            to_combine = new
    if to_combine is not None:
        finals.append(to_combine)

    return finals


def _can_combine(area1, area2):
    _min = min(area1['minPrice'], area2['minPrice'])
    _max = max(area1['maxPrice'], area2['maxPrice'])

    return _max - _min < (area1['maxPrice'] + area2['maxPrice'] - area1['minPrice'] - area2['minPrice']) * 1.02


def _get_combined(area1, area2):
    _min = min(area1['minPrice'], area2['minPrice'])
    _max = max(area1['maxPrice'], area2['maxPrice'])

    return {'minPrice': _min, 'maxPrice': _max, 'index': f"{area1['index']}_{area2['index']}"}


def _prepare_dataframe_volumeprofile(df: pd.DataFrame):
    df["index"] = df.index
    df["valid"] = False


def _set_ignore(upP, downP, df: pd.DataFrame):
    _len = len(df)
    start, end = (
        int(_len / 100 * downP),
        int(_len - _len / 100 * upP),
    )
    df.loc[start:end, "valid"] = True


def _get_requested_volprofile_for_waves(df, pivots, upWaveNums, downWaveNums, nBins):
    df["inVolumeProfile"] = False

    for upWaveNum in upWaveNums:
        waveIndices = _findWave(pivots, upWaveNum, DIRECTION.UP)
        cond1 = np.logical_and(df.index >= waveIndices[0], df.index < waveIndices[1])
        df["inVolumeProfile"] = np.logical_or(cond1, df["inVolumeProfile"])

    for downWaveNum in downWaveNums:
        waveIndices = _findWave(pivots, downWaveNum, DIRECTION.DOWN)
        cond1 = np.logical_and(df.index >= waveIndices[0], df.index < waveIndices[1])
        df["inVolumeProfile"] = np.logical_or(cond1, df["inVolumeProfile"])

    df = df[df.inVolumeProfile == True]

    res = vp.getVP(df, nBins=nBins)
    res["index"] = res.index
    return res



from zigzag import peak_valley_pivots
import pandas as pd
from gstargets.config import DIRECTION, Develop
from gstargets.shared import (
    _get_min_idxs,
    _get_jumped_idxs,
    _get_requested_volprofile_for_waves,
    _set_ignore,
    _prepare_dataframe_volumeprofile,
)


def _get_tp_level_index(
    volprofile_result,
    tradeSide,
    ignorePercentageUp=20,
    ignorePercentageDown=20,
    thresholdType2=0.5,
):
    _prepare_dataframe_volumeprofile(volprofile_result)
    _set_ignore(ignorePercentageUp, ignorePercentageDown, volprofile_result)

    answers = []
    for minIdx in _get_min_idxs(volprofile_result[volprofile_result["valid"]]):
        answers.append({"type": "type1", "index": minIdx})
    for jumpedIdx in _get_jumped_idxs(volprofile_result, thresholdType2, tradeSide):
        answers.append({"type": "type2", "index": jumpedIdx})
    if Develop:
        print(answers)
    return answers


def _convert_index_to_price(vpdf, tpsIdx, trend):
    res = []
    for _, dic in enumerate(tpsIdx):
        price = (
            vpdf.iloc[dic["index"]].minPrice
            if trend == DIRECTION.UP
            else vpdf.iloc[dic["index"]].maxPrice
        )
        res.append({"price": price, "type": dic["type"]})
    if Develop:
        print(res)
    return res


def getTPs(
    df: pd.DataFrame,
    tradeSide,
    thresholdType2=0.5,
    upWaveNums=[],
    downWaveNums=[],
    nBins=20,
    ignorePercentageUp=20,
    ignorePercentageDown=20,
    zigzagUpThreshold=0.3,
    zigzagDownThreshold=-0.3,
    returnVP=False,
    returnPivot=False,
):
    """suggest target points based on wave

    params:
        df: pd.DataFrame -> appropriate for volume profile which I had explained in the volprofile package.
                            Checkout `volprofile.getVP` function.
                        It means that it should have `price` and `volume`
                        Also it must provide the basic ohlcv data.
        tradeSide: str: ['UP', 'DOWN']
        thresholdType2: float -> calculate type2 TPs based on this
        upWaveNums: list[int] -> Exclusive list of the upward waves to be selected for volume profile indicator.
                        It starts from 1 which means the last upward wave except for the current wave.
                        default ([])
        downWaveNums: list[int] -> same as upWaveNums but in downWard direction default ([])
        nBins: int -> needed for volume profile
                        default (20)
        ignorePercentageUp: int -> ignore the results of the volume profile calculation from the top
                        default (20)
        ignorePercentageDown: int -> ignore the results of the volume profile calculation from the bottom
                        default (20)
        zigzagUpThreshold: float -> default 0.3
        zigzagDownThreshold: float -> default 0.3
    return:
        list[Dict{'type', 'price'}]
        types can be either
            type1 : the level with minimum volume
            type2 : the level which is significantly jumped by thresholdType2(ex. 0.5) from last bar
            type3 : before some strong volume level(s)
    """

    df.reset_index(inplace=True, drop=True)

    pivots = peak_valley_pivots(df.close, zigzagUpThreshold, zigzagDownThreshold)

    res = _get_requested_volprofile_for_waves(
        df, pivots, upWaveNums, downWaveNums, nBins
    )
    TPsIdx = _get_tp_level_index(
        res,
        tradeSide,
        ignorePercentageUp,
        ignorePercentageDown,
        thresholdType2=thresholdType2,
    )
    TPs = _convert_index_to_price(res, TPsIdx, tradeSide)
    toReturn = TPs
    # the following line are only for testing purposes
    if returnVP or returnPivot:
        toReturn = {"tps": TPs}
    if returnVP:
        toReturn["vp"] = res
    if returnPivot:
        toReturn["pivots"] = pivots
    return toReturn



from zigzag import peak_valley_pivots
import pandas as pd
from gstargets.config import DIRECTION, Develop
from gstargets.shared import (
    _get_high_volume_levels,
    _get_requested_volprofile_for_waves,
    _set_ignore,
    _findWave,
    _fractional_candle,
    _prepare_dataframe_volumeprofile,
)


def getReversalArea(
    df: pd.DataFrame,
    tradeSide,
    entryPoint=None,
    upWaveNums=[1],
    downWaveNums=[1],
    nBins=20,
    ignorePercentageUp=20,
    ignorePercentageDown=20,
    zigzagUpThreshold=0.3,
    zigzagDownThreshold=-0.3,
    returnVP=False,
    returnPivot=False,
):
    reversalAreas = []
    pivots = peak_valley_pivots(df.close, zigzagUpThreshold, zigzagDownThreshold)
    volprofile_result = _get_requested_volprofile_for_waves(
        df, pivots, upWaveNums, downWaveNums, nBins
    )
    _prepare_dataframe_volumeprofile(volprofile_result)
    _set_ignore(ignorePercentageUp, ignorePercentageDown, volprofile_result)

    res = _get_high_volume_levels(
        volprofile_result,
        current_price=entryPoint,
        trade_side=tradeSide,
    )
    reversalAreas.extend(res)
    _expand_reversal_areas(df, reversalAreas, tradeSide, pivots, zigzagUpThreshold, zigzagDownThreshold)
    if Develop:
        print(reversalAreas)
    toReturn = reversalAreas
    if returnVP or returnPivot:
        toReturn = {"reversal_areas": toReturn}
    if returnVP:
        toReturn["vp"] = volprofile_result
    if returnPivot:
        toReturn["pivots"] = pivots
    return toReturn


def _expand_reversal_areas(df, reversalAreas, tradeSide, pivots, upThr, downThr):
    expand = "maxPrice"
    if tradeSide == DIRECTION.UP:
        expand = "minPrice"

    after, before = _findWave(pivots, waveNum=1, direction=DIRECTION.UP if tradeSide == DIRECTION.DOWN else DIRECTION.DOWN)

    print(after, before)
    # first iteration to get the approximate area
    for _, row in df.iloc[after:before, :].iterrows():
        for j, rev_area in enumerate(reversalAreas):
            _dict = row.to_dict()
            if (
                _fractional_candle(0.5, _dict)
                and _dict["high"] > rev_area[expand]
                and _dict["low"] < rev_area[expand]
            ):
                reversalAreas[j][expand] = (
                    _dict["high"] if tradeSide == DIRECTION.DOWN else _dict["low"]
                )
    # TODO remove hard-code
    micro_pivots = peak_valley_pivots(df.close, upThr/5, downThr/5)

    valid_pivots = set()
    Idx = 1
    while True:
        leftU, _ = _findWave(micro_pivots, Idx, direction=DIRECTION.UP)
        leftD, _ = _findWave(micro_pivots, Idx, direction=DIRECTION.DOWN)

        if leftU < before and leftU > after:
            valid_pivots.add(leftU)
        if leftD < before and leftD > after:
            valid_pivots.add(leftD)

        if leftD < after and leftU < after:
            break
        Idx += 1

    for idx, row in df.iloc[after:before, :].iterrows():
        for j, rev_area in enumerate(reversalAreas):
            _dict = row.to_dict()
            if (
                abs(_dict['close'] - rev_area[expand]) / rev_area[expand] < 0.13
                and idx in valid_pivots
            ):
                reversalAreas[j][expand] = (
                    _dict["high"] if tradeSide == DIRECTION.DOWN else _dict["low"]
                )



import plotly.graph_objects as go
import numpy as np
import pandas as pd
from plotly.subplots import make_subplots

from gstargets.target_point import getTPs
from gstargets.reversal import getReversalArea
from gstargets.shared import _findWave
from gstargets.config import DIRECTION


def plot_target_points(
    df: pd.DataFrame,
    tradeSide,
    maximumAcceptableBarType3=0,
    thresholdType2=0.5,
    entryPoint=None,
    upWaveNums=[],
    downWaveNums=[],
    nBins=20,
    windowType3=2,
    ignorePercentageUp=20,
    ignorePercentageDown=20,
    zigzagUpThreshold=0.3,
    zigzagDownThreshold=-0.3,
):
    """completely identical inputs to getTPs function
    you can use this function to display the results.
    """
    df.reset_index(inplace=True, drop=True)
    forPlot = df.copy()

    res = getTPs(
        df,
        tradeSide,
        thresholdType2=thresholdType2,
        upWaveNums=upWaveNums,
        downWaveNums=downWaveNums,
        nBins=nBins,
        ignorePercentageUp=ignorePercentageUp,
        ignorePercentageDown=ignorePercentageDown,
        zigzagUpThreshold=zigzagUpThreshold,
        zigzagDownThreshold=zigzagDownThreshold,
        returnPivot=True,
        returnVP=True,
    )
    TPs, res, pivots = res["tps"], res["vp"], res["pivots"]

    fig = make_subplots(rows=1, cols=2)
    fig.add_trace(
        go.Bar(
            x=res.aggregateVolume, y=(res.minPrice + res.maxPrice) / 2, orientation="h"
        ),
        row=1,
        col=2,
    )

    close = forPlot.close
    fig.add_trace(
        go.Scatter(name="close", y=close, mode="lines", marker_color="#D2691E")
    )
    for waveNum in upWaveNums:
        wave = _findWave(pivots, waveNum, DIRECTION.UP)
        fig.add_trace(
            go.Scatter(
                name="close",
                x=np.arange(len(close))[wave[0] : wave[1]],
                y=close[wave[0] : wave[1]],
                mode="lines",
                marker_color="black",
            )
        )
    for waveNum in downWaveNums:
        wave = _findWave(pivots, waveNum, DIRECTION.DOWN)
        fig.add_trace(
            go.Scatter(
                name="close",
                x=np.arange(len(close))[wave[0] : wave[1]],
                y=close[wave[0] : wave[1]],
                mode="lines",
                marker_color="black",
            )
        )
    fig.add_trace(
        go.Scatter(
            name="top",
            x=np.arange(len(close))[pivots == 1],
            y=close[pivots == 1],
            mode="markers",
            marker_color="green",
        )
    )
    fig.add_trace(
        go.Scatter(
            name="bottom",
            x=np.arange(len(close))[pivots == -1],
            y=close[pivots == -1],
            mode="markers",
            marker_color="red",
        )
    )

    for line in TPs:
        fig.add_hline(
            y=line["price"], line_width=3, line_dash="dash", line_color="green"
        )

    fig.show()


def _manual_test_tp():
    # TODO : use yfinance and pytse-client to get plenty of tickers
    # and test automatically
    path = "~/Downloads/data/tickers_data/test.csv"
    df = pd.read_csv(path)

    nBins = 20
    tradeSide = DIRECTION.UP
    downWaveNums = [1]
    zigzagDownThreshold = -0.3
    zigzagUpThreshold = 0.3

    n = 1000
    df = df[-n:]

    df["price"] = (df["high"] + df["low"]) / 2
    df = df[["volume", "price", "close"]]
    plot_target_points(
        df,
        tradeSide,
        nBins=nBins,
        downWaveNums=downWaveNums,
        zigzagUpThreshold=zigzagUpThreshold,
        zigzagDownThreshold=zigzagDownThreshold,
    )


def plot_reversal_area(df, tradeSide, upWaveNums=[], downWaveNums=[]):
    df.reset_index(inplace=True, drop=True)
    forPlot = df.copy()

    res = getReversalArea(
        df,
        tradeSide,
        zigzagUpThreshold=0.4,
        zigzagDownThreshold=-0.4,
        upWaveNums=upWaveNums,
        downWaveNums=downWaveNums,
        returnPivot=True,
        returnVP=True,
    )

    rev_areas, res, pivots = res["reversal_areas"], res["vp"], res["pivots"]

    fig = make_subplots(rows=1, cols=2)
    fig.add_trace(
        go.Bar(
            x=res.aggregateVolume, y=(res.minPrice + res.maxPrice) / 2, orientation="h"
        ),
        row=1,
        col=2,
    )

    close = forPlot.close
    fig.add_trace(
        go.Scatter(name="close", y=close, mode="lines", marker_color="#D2691E")
    )
    for waveNum in upWaveNums:
        wave = _findWave(pivots, waveNum, DIRECTION.UP)
        fig.add_trace(
            go.Scatter(
                name="close",
                x=np.arange(len(close))[wave[0] : wave[1]],
                y=close[wave[0] : wave[1]],
                mode="lines",
                marker_color="black",
            )
        )
    for waveNum in downWaveNums:
        wave = _findWave(pivots, waveNum, DIRECTION.DOWN)
        fig.add_trace(
            go.Scatter(
                name="close",
                x=np.arange(len(close))[wave[0] : wave[1]],
                y=close[wave[0] : wave[1]],
                mode="lines",
                marker_color="black",
            )
        )
    fig.add_trace(
        go.Scatter(
            name="top",
            x=np.arange(len(close))[pivots == 1],
            y=close[pivots == 1],
            mode="markers",
            marker_color="green",
        )
    )
    fig.add_trace(
        go.Scatter(
            name="bottom",
            x=np.arange(len(close))[pivots == -1],
            y=close[pivots == -1],
            mode="markers",
            marker_color="red",
        )
    )

    for line in rev_areas:
        fig.add_hline(
            y=line["maxPrice"], line_width=3, line_dash="dash", line_color="green"
        )
        fig.add_hline(
            y=line["minPrice"], line_width=3, line_dash="dash", line_color="blue"
        )

    fig.show()


def _manual_test_reversal():
    path = "~/Downloads/data/tickers_data/test.csv"
    df = pd.read_csv(path)

    tradeSide = DIRECTION.UP
    upWaveNums = []
    downWaveNums = [1]

    n = 1000
    df = df[-n:-n//3]

    df["price"] = (df["high"] + df["low"]) / 2
    plot_reversal_area(df, tradeSide, upWaveNums=upWaveNums, downWaveNums=downWaveNums)


if __name__ == "__main__":
    _manual_test_tp()
    _manual_test_reversal()
